# README for Python Project

## Description

This project is a Python implementation of the Passenger-Oriented Flight Rescheduling Problem (PFRSP). The project is divided into several scripts, each serving a specific purpose. The primary functionalities include solving optimization problems related to flight schedules, considering various constraints.

## Installation required
- **Gurobi:** The script relies on the Gurobi optimization solver. Ensure Gurobi is installed before running the script. You can install Gurobi by following the instructions on the [Gurobi website](https://www.gurobi.com/).


## Scripts

1. **constants.py**
   - Contains constant values and configurations used throughout the project.

2. **pre_processing.py**
   - Handles pre-processing tasks, such as  reading input files, creating the instances, cropping the instances and computing terminal occupancies.

3. **post_processing.py**
   - Manages post-processing tasks, including computing Performance Indicators and displaying results.

4. **ilp_solving.py**
   - Implements a direct solving approach through Gurobi ILP solver to solve the PFRSP on one time window.

5. **heuristic_solving.py**
   - Implements a heuristic approach for solving flight scheduling problems.

6. **timeWindowManagement.py**
   - Manages sliding time window and flight status.

7. **main.py**
   - The main script that need to be run, enscompass the whole run of the sliding time window to solve a sequence of PFRSP instances and reschedule flights between a stating and an ending time. 

8. **airport.py**
   - Defines the `Airport` class, which represents the airport structure.

9. **passenger.py**
   - Implements the `Passenger` class, handling passenger-related features.

## Usage

To run the project, execute `main.py` with the desired parameters. The script reads an airport instance, solves a sequence of PFRSP instances, and performs post-processing. Various options, such as using ILP or heuristics, reading initial solutions, and saving figures, can be configured.

Example:
```bash
python main.py --greedy=False --read_sol=False --savefig=False --filename='Results/flightset_results.txt'

