import matplotlib.pyplot as plt
import numpy as np
from airport import *
from flight import *
from passenger import *
import pandas as pd
from pre_processing import crop_instance
from timeWindowManagement import set_status_flights

def update_flightset(flightset, flights_a, flights_d):  # ajoute information flights time window apres optimisation  au flightset entier
    for key in flights_a:
        flightset.flights_a[key] = flights_a[key]
    for key in flights_d:
        flightset.flights_d[key] = flights_d[key]
    return flightset


def write_solution(flightset, label):
    filename = FOLDER_RESULTS + f'flightset_results_{SCENARIO}.txt'
    with open(filename, 'w') as file:
        file.write(f"NbArrivalFlights: {len(flightset.flights_a)} \n")
        for key in flightset.flights_a:
            file.write(flightset.flights_a[key].to_string() + '\n')
        file.write(f"NbDepartureFlights: {len(flightset.flights_d)} \n")
        for key in flightset.flights_d:
            file.write(flightset.flights_d[key].to_string() + '\n')





def compute_throughput_runway(airport, flightset):  # time window to compute throughput
    TIME_SCOPE = (25 * 60 * 60)//TIME_STEP
    flights_a = flightset.flights_a
    flights_d = flightset.flights_d
    runways_bo, runways_ao = np.zeros((4, TIME_SCOPE)), np.zeros((4, TIME_SCOPE))
    for key in flights_a:
        if flights_a[key].status not in [-1]:  # not planned, used if simulation stop before T_MAX= 24h-TW
            init_landing_time = int(flights_a[key].landing_time)
            final_landing_time = int(flights_a[key].final_landing_time)
            init_runway = flights_a[key].runway
            final_runway = flights_a[key].final_runway
            runways_bo[init_runway][init_landing_time] += 1
            runways_ao[final_runway][final_landing_time] += 1

    for key in flights_d:
        if flights_d[key].status not in [-1]:  # not planned, used if simulation stop before T_MAX= 24h-TW
            terminal = int(flights_d[key].terminal)
            init_obt = int(flights_d[key].obt)
            final_obt = int(flights_d[key].final_obt)
            init_runway = flights_d[key].runway
            final_runway = flights_d[key].final_runway
            init_takeoff = int(init_obt + airport.taxi_out_duration[terminal][init_runway])
            final_takeoff = int(final_obt + airport.taxi_out_duration[terminal][final_runway])
            runways_bo[init_runway][init_takeoff] += 1
            runways_ao[final_runway][final_takeoff] += 1
    return runways_bo,runways_ao


def display_throughput_constraints(airport, runways_bo, runways_ao, is_step_constraint, savefig=False):
    fig, axs = plt.subplots(4, figsize=(10, 6))
    plt.subplots_adjust(left=0.125,
                        bottom=0.1,
                        right=0.9,
                        top=0.9,
                        wspace=0.2,
                        hspace=0.45)
    for i in range(NB_RUNWAYS):
        axs[i].step(np.arange(len(runways_bo[i])), runways_bo[i],
                    label=f'Initial ({airport.runways[i].name})',where='post')
        axs[i].step(np.arange(len(runways_ao[i])), runways_ao[i],
                    label=f'Optimized ({airport.runways[i].name})',where='post')
        axs[i].set_ylabel('Number of \n movements', fontsize=12)
        if is_step_constraint == True:
            x = np.arange(T_MIN,T_MAX+WINDOW_DURATION,CAPACITY_WINDOW)
            Y = list(airport.runways[i].capacities.values())
            Y_step = Y
            axs[i].step(x, Y_step, c='r', linestyle='--',where='post')
        else:
            axs[i].axhline(airport.runways[i].capacity, c='r', linestyle='--')
        axs[i].legend(loc='upper right')
        axs[i].set_xticks(np.arange(0, T_MAX + 60*60//TIME_STEP, 60*60//TIME_STEP))
        axs[i].set_xticklabels(["{0}h".format(e) for e in range(0, 25)], rotation=45)
        axs[i].set_xlim(START_MODE_DISRUPTION_TIME,T_MAX)
    if savefig:
        plt.savefig(FOLDER_RESULTS + f'throughput_constraints_{SCENARIO}.png', bbox_inches='tight')
    else:
        plt.show()


def compute_congestion(airport, flightset):  # time window to compute throughput
    flights_a = flightset.flights_a
    flights_d = flightset.flights_d
    TIME_SCOPE = (T_MAX - T_MIN +WINDOW_DURATION)
    init_terminals_congestion = np.zeros((NB_TERMINALS, TIME_SCOPE))
    final_terminals_congestion = np.zeros((NB_TERMINALS, TIME_SCOPE))
    for k in range(NB_TERMINALS):
        init_terminals_congestion[k][:] = airport.terminals[k].nb_initial_departures
        final_terminals_congestion[k][:] = airport.terminals[k].nb_initial_departures
    init_taxi_congestion = np.zeros(TIME_SCOPE)
    final_taxi_congestion = np.zeros(TIME_SCOPE)
    for key in flights_a:
        terminal = flights_a[key].terminal
        init_landing_time = round(flights_a[key].landing_time)
        final_landing_time = round(flights_a[key].final_landing_time)
        init_ibt = round(flights_a[key].ibt)
        final_ibt = round(flights_a[key].final_ibt)
        init_terminals_congestion[terminal][init_ibt+1:] += 1
        init_taxi_congestion[init_landing_time+1:init_ibt+1] += 1
        final_terminals_congestion[terminal][final_ibt+1:] += 1
        final_taxi_congestion[final_landing_time+1:final_ibt+1] += 1
    for key in flights_d:
        terminal = flights_d[key].terminal
        init_obt = round(flights_d[key].obt)
        final_obt = round(flights_d[key].final_obt)
        idx_init_runway_out = flights_d[key].runway
        idx_final_runway_out = flights_d[key].final_runway
        init_takeoff_time = round(init_obt + airport.taxi_out_duration[terminal][idx_init_runway_out])
        final_takeoff_time = round(final_obt + airport.taxi_out_duration[terminal][idx_final_runway_out])
        init_terminals_congestion[terminal][init_obt+1:] -= 1
        init_taxi_congestion[init_obt+1:init_takeoff_time+1] += 1
        final_terminals_congestion[terminal][final_obt+1:] -= 1
        final_taxi_congestion[final_obt+1:final_takeoff_time+1] += 1
    return init_terminals_congestion, final_terminals_congestion, init_taxi_congestion, final_taxi_congestion


def display_capacity_constraint(airport, init_terminals_congestion, final_terminals_congestion, init_taxi_congestion,
                                final_taxi_congestion, is_taxi_step_constraint, is_terminal_step_constraint,
                               savefig=False):
    subset_terminals = ['T2E','T2F']
    nb_terminals = len(subset_terminals)
    fig, ax = plt.subplots(nb_terminals+1, figsize=(12, 6))
    plt.subplots_adjust(left=0.125,
                        bottom=0.1,
                        right=0.9,
                        top=0.9,
                        wspace=0.2,
                        hspace=0.6)
    x_capacity = np.arange(T_MIN, T_MAX+WINDOW_DURATION, CAPACITY_WINDOW)
    X = np.arange(T_MIN, T_MAX+WINDOW_DURATION)
    for i,terminal in enumerate(subset_terminals):
        k=0
        for j in range(NB_TERMINALS):
            if airport.terminals[j].terminal_name == terminal:
                k =j
                break
        ax[i].step(X,init_terminals_congestion[k], label=f"Initial occupancy {terminal}",where='post')
        ax[i].step(X,final_terminals_congestion[k], label=f"Optimized occupancy {terminal}",where='post')
        if is_terminal_step_constraint:
            Y_k = list(airport.terminals[k].capacities.values())
            Y_k_step = Y_k # + [Y_k[-1]]
            ax[i].step(x_capacity, Y_k_step, c='r', linestyle='--',where='post')
        else:
            ax[i].axhline(airport.terminals[k].capacity, c='r', linestyle='--')
        ax[i].set_xticks(np.arange(0, T_MAX + 60*60//TIME_STEP, 60*60//TIME_STEP))
        ax[i].set_xticklabels(["{0}h".format(e) for e in range(0, 25)], rotation=45)
        ax[i].legend(loc=1)
        ax[i].set_xlim(START_MODE_DISRUPTION_TIME, T_MAX)
    ax[nb_terminals].step(X,init_taxi_congestion, label=f'initial taxi occupancy', where='post')
    ax[nb_terminals].step(X,final_taxi_congestion, label=f'optimized taxi occupancy',where='post')
    if is_taxi_step_constraint:
        Y_taxi = list(airport.taxi.capacities.values())
        Y_taxi_step = Y_taxi # + [Y_taxi[-1]]
        ax[nb_terminals].step(x_capacity, Y_taxi_step, c='r', linestyle='--',where='post')
    else:
        ax[nb_terminals].axhline(airport.taxi.capacity, c='r', linestyle='--')  # ,label='{0} capacity'.format(name))
    ax[nb_terminals].set_xticks(np.arange(0, T_MAX + 60*60//TIME_STEP, 60*60//TIME_STEP))
    ax[nb_terminals].set_xticklabels(["{0}h".format(e) for e in range(0, 25)], rotation=45)
    ax[nb_terminals].legend(loc=1)
    ax[nb_terminals].set_xlim(START_MODE_DISRUPTION_TIME, T_MAX)

    if savefig:
        plt.savefig( FOLDER_RESULTS + f'capacities_constraints_{SCENARIO}.png', bbox_inches='tight')
    else:
        plt.show()


def compute_missed_passengers(flightset, passenger):
    flights_d = flightset.flights_d
    t_start = START_AIRPORT_DISRUPTION_TIME
    t_end = END_AIRPORT_DISRUPTION_TIME
    init_missed_pax = np.zeros((t_end-t_start))
    final_missed_pax = np.zeros((t_end - t_start))
    for key in flights_d:
        init_takeoff = int(flights_d[key].takeoff)
        if (init_takeoff >= t_start) and (init_takeoff < t_end):
            init_obt = int(flights_d[key].obt)
            final_obt = int(flights_d[key].final_obt)
            init_missed_pax_i = passenger.ground_pax_connected_table[key][0]
            final_missed_pax_i = passenger.ground_pax_connected_table[key][final_obt - init_obt]

            init_missed_pax[init_takeoff-START_AIRPORT_DISRUPTION_TIME] += init_missed_pax_i
            final_missed_pax[init_takeoff-START_AIRPORT_DISRUPTION_TIME] += final_missed_pax_i
    return init_missed_pax, final_missed_pax


def display_cumul_passengers(init_missed_pax, final_missed_pax, ax):
    x = np.arange(T_MIN,T_MAX)
    x = x * (TIME_STEP/(60*60))
    y_init_cumul = np.cumsum(init_missed_pax)
    y_final_cumul = np.cumsum(final_missed_pax)
    ax.plot(x,y_init_cumul, label='Initial schedule', alpha=0.5)
    ax.plot(x,y_final_cumul, label='Optimized schedule', alpha=0.5)
    ax.set_xticks(np.arange(24))
    ax.set_ylabel('Cumulative number of stranded passengers', fontsize=12)
    ax.set_xticklabels(["{0}h".format(e) for e in range(0, 24)], rotation=45)
    ax.legend(loc=2)



def display_missed_passengers(init_missed_pax, final_missed_pax, ax):
    t_start,t_end  = START_AIRPORT_DISRUPTION_TIME, END_AIRPORT_DISRUPTION_TIME
    step_hour = 60*60//TIME_STEP
    x = np.arange(t_start,t_end,step_hour)
    y_init = [np.sum(init_missed_pax[i:i+step_hour]) for i in range(0,t_end-t_start,step_hour)]
    y_final = [np.sum(final_missed_pax[i:i+step_hour]) for i in range(0,t_end-t_start,step_hour)]
    ax.bar(x-2, np.array(y_init), width=4,label='Initial schedule', alpha=1)
    ax.bar(x+2, np.array(y_final), width=4,label='Optimized schedule', alpha=0.7)
    ax.set_xticks(np.arange(t_start,T_MAX,step_hour))
    ax.set_ylabel('Number of stranded \n passengers', fontsize=12)
    ax.set_xticklabels(["{0}h".format(e) for e in range(t_start*TIME_STEP//(60*60), T_MAX*TIME_STEP//(60*60))], rotation=45)
    ax.legend()
    ax.set_xlim(t_start-step_hour,t_end)


def compute_delay_disruption(flightset):
    t_start = START_AIRPORT_DISRUPTION_TIME
    t_end = END_AIRPORT_DISRUPTION_TIME
    # t_end = T_MAX
    delays_landing = {}
    delays_takeoff = {}
    total_delay_landing = 0
    total_delay_takeoff = 0
    nb_slots_violated = 0
    induced_delay = 0
    cpt_earlier_flights = 0
    for flight in flightset.flights_a.values():
        if (flight.landing_time >= t_start) and (flight.landing_time < t_end) :
            arrival_delay = flight.final_landing_time - flight.landing_time
            if arrival_delay <0:
                cpt_earlier_flights +=1
            delay = np.abs(arrival_delay) * TIME_STEP//60
            delays_landing[flight.id] = delay# in minutes
            total_delay_landing += delay
        elif flight.landing_time > t_end:
            induced_delay += np.abs(flight.final_landing_time - flight.landing_time) * TIME_STEP//60
    for flight in flightset.flights_d.values():
        if (flight.takeoff >= t_start) and (flight.takeoff < t_end) :
            delay = np.abs(flight.final_takeoff - flight.takeoff)
            if delay > DELTA_SLOT:
                nb_slots_violated += 1
            delay = delay * TIME_STEP//60
            delays_takeoff[flight.id] = delay # in minutes
            total_delay_takeoff += delay
        elif flight.takeoff > t_end:
            induced_delay += np.abs(flight.final_takeoff - flight.takeoff) * TIME_STEP//60
    print(f"**** Nb arrival flights landed before scheduled : {cpt_earlier_flights}")
    return delays_landing, delays_takeoff, total_delay_landing, total_delay_takeoff, nb_slots_violated,induced_delay



def display_histogram_delay(flightset,delay_landing, delay_takeoff,  ax):
    hours = np.arange(0, 24)
    y_landing = np.zeros((T_MAX-T_MIN)*TIME_STEP//(60*60))
    nb_landings =  np.zeros((T_MAX-T_MIN)*TIME_STEP//(60*60))
    y_takeoff =  np.zeros((T_MAX-T_MIN)*TIME_STEP//(60*60))
    nb_takeoffs =  np.zeros((T_MAX-T_MIN)*TIME_STEP//(60*60))
    for key in delay_landing:
        h = int(flightset.flights_a[key].landing_time*TIME_STEP//(60*60))
        y_landing[h] += delay_landing[key]
        nb_landings[h] += 1
    for key in delay_takeoff:
        h = int(flightset.flights_d[key].takeoff*TIME_STEP//(60*60))
        y_takeoff[h] += delay_takeoff[key]
        nb_takeoffs[h] += 1
    for i in range(len(y_landing)):
        if nb_landings[i] != 0:
            y_landing[i] /= nb_landings[i]
    for i in range(len(y_takeoff)):
        if  nb_takeoffs[i] != 0:
            y_takeoff[i] /= nb_takeoffs[i]
    ax.step(hours+0.5, y_takeoff, alpha=1, label='Departing flights', color='C2')
    ax.step(hours+0.5, y_landing, alpha=1, label='Arriving flights', color='C3', linestyle='--')

    ax.set_ylabel('Average aircraft \n delay (min)', fontsize=12)
    ax.legend()
    ax.set_xticks(hours)
    ax.set_xticklabels(["{0}h".format(e) for e in range(0, 24)], rotation=45)
    step_hour = TIME_STEP/(60*60)
    ax.set_xlim(int(START_AIRPORT_DISRUPTION_TIME*step_hour)-0.9,int(END_AIRPORT_DISRUPTION_TIME*step_hour))
    ax.set_ylim(0,int(max(y_takeoff))+1)


def display_barplot_delay(flightset,savefig):
    fig,ax = plt.subplots(figsize=(10,4))
    x_min = -1
    X = np.arange(x_min,DELTA_OBT+1)
    Y_arrival = np.zeros(DELTA_OBT-x_min+1)
    Y_departure = np.zeros(DELTA_OBT-x_min+1)
    t_start = START_AIRPORT_DISRUPTION_TIME
    t_end = END_AIRPORT_DISRUPTION_TIME
    for flight in flightset.flights_a.values():
        if (flight.landing_time >= t_start) and (flight.landing_time < t_end):
            arrival_delay = flight.final_landing_time - flight.landing_time
            Y_arrival[arrival_delay-x_min] +=1
    for flight in flightset.flights_d.values():
        if (flight.takeoff >= t_start) and (flight.takeoff < t_end):
            departure_delay = flight.final_takeoff - flight.takeoff
            if departure_delay<DELTA_OBT-x_min:
                Y_departure[departure_delay-x_min] +=1
    width =0.3
    ax.bar(X+ width/2, Y_departure, alpha=1, width=width, label='Departing flights', color='C0', linestyle='--')
    ax.bar(X- width/2, Y_arrival, alpha=1, width=width, label='Arriving flights', color='C1', linestyle='--')
    ax.set_xlabel('Delay', fontsize=12)
    ax.set_xticks(X)
    ax.set_ylim(0,350)
    ax.set_xticklabels([f'{e*5}min' for e in X], rotation=45)
    ax.set_ylabel('Number of flights', fontsize=12)
    ax.legend()
    Y_proportion =  100*Y_departure/np.sum(Y_departure)
    print("pct on time:", [f"{X[i]}:{Y_proportion[i]}" for i in range(len(X))])
    print("delay arrivals:", [f"{X[i]}:{Y_arrival[i]}" for i in range(len(X))])

    if savefig:
        plt.savefig( FOLDER_RESULTS + f'barplot_flight_volume_per_delay_{SCENARIO}.png', bbox_inches='tight')
    else:
        plt.show()




def display_buffer_transfer_air_pax(flightset, passengers,savefig):
    init_transfers=[]
    optim_transfers=[]
    total_connecting_pax = 0
    for key in passengers.air_connected_pax_table:
        (id_dep_flight, id_arr_flight) = key
        if flightset.flights_d[id_dep_flight].status != -1:
            min_connection_time = passengers.air_connected_pax_table[key].min_connection_time
            nb_connecting_pax = passengers.air_connected_pax_table[key].nb_pax
            total_connecting_pax += nb_connecting_pax
            init_ibt = flightset.flights_a[id_arr_flight].ibt
            init_obt = flightset.flights_d[id_dep_flight].obt
            init_transfers += [(init_obt-init_ibt-min_connection_time)*TIME_STEP/60 for _ in range(nb_connecting_pax)]
            ibt = flightset.flights_a[id_arr_flight].final_ibt
            obt = flightset.flights_d[id_dep_flight].final_obt
            optim_transfers += [(obt-ibt-min_connection_time)*TIME_STEP/60 for _ in range(nb_connecting_pax)]
            if obt - ibt < min_connection_time:
                res = False
                print(obt - ibt, '  ', min_connection_time)
    fig,axes = plt.subplots(2,figsize=(10,6))
    axes[0].hist(init_transfers,color='r',alpha=0.8,label='Initial schedule',bins=20)
    axes[1].hist(optim_transfers,color='b',alpha=0.5,label='Optimzed schedule',bins=20)
    xticks = [i for i in range(0,150,15)]
    axes[1].set_xlabel('Buffer time for passenger transfers (minutes)')
    for i in range(2):
        axes[i].set_xticks(xticks)
        axes[i].set_xticklabels([f'{i}' for i in xticks])
        axes[i].set_ylabel('Number of passengers')
        axes[i].legend()
    print('nb connecting pax: ', total_connecting_pax)
    print('nb connecting flight pairs:', len(passengers.air_connected_pax_table))
    print(f'average connecting time: initial:{np.mean(init_transfers)}  after rescheduling: {np.mean(optim_transfers)}')
    optim_transfers = np.array(optim_transfers)
    print('optim buffer transfer >2hours:', len(optim_transfers[optim_transfers>120]))
    print('optim buffer transfer =0min:', len(optim_transfers[optim_transfers==0]))
    if savefig:
        plt.savefig(FOLDER_RESULTS + f'buffer_transfer_air_pax_{SCENARIO}.png', bbox_inches='tight')

    else:
        plt.show()

def display_slot_violation(flightset,savefig):
    fig,ax = plt.subplots(figsize=(10,4))
    X = np.arange(24)
    nb_flights = np.zeros(24)
    Y = np.zeros(24)
    for flight in flightset.flights_d.values():
        if np.abs(flight.final_takeoff-flight.takeoff) > DELTA_SLOT:
            Y[int(flight.takeoff * TIME_STEP//(60*60))] +=1
        nb_flights[int(flight.takeoff * TIME_STEP//(60*60))] +=1
    ax.bar(X,nb_flights-Y,label='Satisfied slots')

    ax.bar(X,Y,bottom=nb_flights-Y,label='Violated slots')
    ax.legend()
    ax.set_ylabel('Total number of slots')
    xticks = np.arange((START_AIRPORT_DISRUPTION_TIME*TIME_STEP)//3600, (END_AIRPORT_DISRUPTION_TIME*TIME_STEP)//3600+1)
    ax.set_xticks(xticks)
    ax.set_xticklabels(["{0}h".format(e) for e in xticks], rotation=45)
    ax.set_xlim((START_AIRPORT_DISRUPTION_TIME*TIME_STEP)//3600-0.5, (END_AIRPORT_DISRUPTION_TIME*TIME_STEP)//3600-0.5)
    if savefig:
        plt.savefig( FOLDER_RESULTS + f'slots_violation_histogram_{SCENARIO}.png', bbox_inches='tight')
    else:
        plt.show()


def display_hourly_stranded_flight_delay(init_missed_pax,final_missed_pax,flightset,delay_landing,delay_takeoff,savefig):
    fig, axes = plt.subplots(2, figsize=(8, 5))
    plt.subplots_adjust(left=0.125,
                        bottom=0.1,
                        right=0.9,
                        top=0.9,
                        wspace=0.3,
                        hspace=0.4)
    display_missed_passengers(init_missed_pax, final_missed_pax, ax=axes[0])
    display_histogram_delay(flightset, delay_landing, delay_takeoff, ax=axes[1])
    if savefig:
        plt.savefig(FOLDER_RESULTS + f'histograms_{SCENARIO}.png',bbox_inches='tight')
    else:
        plt.show()

def is_slot_constraint_satisfied(flightset):
    res = True
    for flight in flightset.flights_d.values():
        init_takeoff = flight.takeoff
        final_takeoff = flight.final_takeoff
        if flight.priority and ((final_takeoff - init_takeoff) > DELTA_SLOT):
            print(flight.priority)
            print(f'**** slot violation for flight {flight.id} init takeoff: {flight.takeoff}  final_takeoff: {final_takeoff}')
            res = False
            break
    print(f'Slot constraint satisfied: {res}')

    return res


def is_turn_around_constraint_satisfied(flightset, airport):
    res = True
    for key in flightset.id_flights_ad:
        flight_arr = flightset.flights_a[key]
        flight_dep = flightset.flights_d[key]
        if flight_dep.status != 1:
            turn_around = flight_dep.final_obt - flight_arr.final_landing_time + \
                          airport.taxi_in_duration[flight_arr.final_runway-1][flight_arr.terminal]
            if turn_around < MIN_TURNAROUND_TIME:
                res = False
                print(f'turn around violated on flight {key} with turn around time = {turn_around}min')
    print(f"Is turn around constraint  satisfied: {res} ")
    return res


def is_air_connecting_pax_satisfied(flightset, passengers):
    res = True
    for key in passengers.air_connected_pax_table:
        (id_dep_flight, id_arr_flight) = key
        if flightset.flights_d[id_dep_flight].status != -1:
            ibt = flightset.flights_a[id_arr_flight].final_ibt
            obt = flightset.flights_d[id_dep_flight].final_obt
            if obt - ibt < passengers.air_connected_pax_table[key].min_connection_time:
                res = False
                print(obt - ibt, '  ', passengers.air_connected_pax_table[key].min_connection_time)
    print(f"Is air connecting pax constraint satisfied: {res}")
    return res





def is_runway_throughput_satisfied(airport, throughputs_ao, is_step_constraint):
    res = [True for _ in range(NB_RUNWAYS)]
    for r in range(NB_RUNWAYS):
        runway_constraint = True
        if is_step_constraint:
            for t in range(T_MIN, T_MAX, CAPACITY_WINDOW):
                if max(throughputs_ao[r][t:t + CAPACITY_WINDOW]) > airport.runways[r].capacities[t//CAPACITY_WINDOW]:
                    runway_constraint = False
                    print(airport.runways[r].name, "  violation at ", round(t*TIME_STEP/3600,2),'h')
                    break
        else:
            if max(throughputs_ao[r]) > airport.runways[r].capacity:
                runway_constraint = False
        res[r] = runway_constraint
    if res == [True for _ in range(NB_RUNWAYS)]:
        print(f'Is runway throughput constraint  satisfied: True')
    else:
        print(f'Is runway throughput constraint  satisfied: False')

    return res


def is_taxi_capacity_satisfied(airport, final_taxi_congestion, is_step_constraint):
    taxi_constraint = True
    if is_step_constraint:
        for t in range(T_MIN, T_MAX, CAPACITY_WINDOW):
            if max(final_taxi_congestion[t:t + CAPACITY_WINDOW]) > airport.taxi.capacities[t//CAPACITY_WINDOW]:
                taxi_constraint = False
                print(f"Taxi violation at  {t*TIME_STEP/(60*60)}h")
                break
    else:
        if max(final_taxi_congestion) > airport.taxi.capacity:
            taxi_constraint = False
    print(f"Is  taxi capacity constraint satisfied : {taxi_constraint}")
    return taxi_constraint


def is_terminal_capacity_satisfied(airport, final_terminals_congestion, is_step_constraint):
    res = [True for _ in range(NB_TERMINALS)]
    for i in range(NB_TERMINALS):
        is_satisfied = True
        if is_step_constraint:
            for t in range(T_MIN, T_MAX, CAPACITY_WINDOW):
                if max(final_terminals_congestion[i][t:t + CAPACITY_WINDOW]) > airport.terminals[i].capacities[t//CAPACITY_WINDOW]:
                    is_satisfied = False
                    print(f"Terminal {airport.terminals[i].terminal_name} violation at {t*TIME_STEP/(60*60)}h")
                    break
        else:
            if max(final_terminals_congestion[i]) > airport.terminals[i].capacity:
                is_satisfied = False
        res[i] = is_satisfied
    if res ==[True for _ in range(NB_TERMINALS)]:
        print(f"Terminal capacity constraint satisfied : True ")
    else:
        print(f"Terminal capacity constraint satisfied : False ")
    return res






def post_processing(airport, flightset, passengers, alpha,beta, test_constraint, display_constraints, savefig,
                    runway_step_constraint, taxi_step_constraint, terminal_step_constraint,is_greedy):
    init_missed_pax, final_missed_pax = compute_missed_passengers(flightset, passengers)
    init_terminals_congestion, final_terminals_congestion, init_taxi_congestion, final_taxi_congestion = compute_congestion(
        airport, flightset)
    runway_bo, runway_ao = compute_throughput_runway(airport, flightset)
    delay_landing, delay_takeoff, total_landing_delay, total_takeoff_delay, nb_slots_violated, induced_delay = compute_delay_disruption(flightset)

    set_status_flights(flightset, airport,START_AIRPORT_DISRUPTION_TIME, END_AIRPORT_DISRUPTION_TIME)
    flightset_crop, passengers_crop, flightset_og, flightset_planned = crop_instance(airport, flightset, passengers,
                                                                          END_AIRPORT_DISRUPTION_TIME)
    nb_dep_flights = len(flightset.flights_d)
    nb_arr_flights = len(flightset.flights_a)
    nb_dep_flights_disruption = len(flightset_crop.flights_d)
    nb_arr_flights_disruption = len(flightset_crop.flights_a)
    objectif = np.sum(final_missed_pax) + ALPHA*(total_landing_delay + total_takeoff_delay) + BETA*nb_slots_violated
    if is_greedy:
        filename = FOLDER_RESULTS + f'summary_results_{SCENARIO}_greedy_epsilon_{EPSILON_HEURISTIC}.txt' #_TW{WINDOW_DURATION_HOUR}h.txt'
    else:
        filename = FOLDER_RESULTS + f'summary_results_{SCENARIO}_alpha_{alpha}_beta_{beta}_no_arrival_deviation.txt' #_delayOBT_{DELTA_OBT*TIME_STEP//60}min.txt'

    with open(filename, 'w') as file:
        nb_stranded_pax_bo = int(sum(init_missed_pax))
        nb_stranded_pax_ao = int(sum(final_missed_pax))
        ratio = (nb_stranded_pax_bo - nb_stranded_pax_ao) /nb_stranded_pax_bo
        str_stranded_pax_bo = f'Number of stranded passengers before optim: {int(sum(init_missed_pax))}'
        str_stranded_pax_ao = f'Number of stranded passengers after optim: {int(sum(final_missed_pax))}'
        print(str_stranded_pax_bo)
        print(str_stranded_pax_ao)
        file.write(str_stranded_pax_bo + '\n')
        file.write(str_stranded_pax_ao + '\n')
        str_nb_dep_flights = f'Nb departure flights: Day= {nb_dep_flights} , during the disruption= {nb_dep_flights_disruption}'
        str_nb_arr_flights = f'Nb arrival flights: Day= {nb_arr_flights} , during the disruption= {nb_arr_flights_disruption}'
        print(str_nb_dep_flights)
        print(str_nb_arr_flights)
        file.write(str_nb_dep_flights + "\n")
        file.write(str_nb_arr_flights + "\n")
        str_stranded_pax_ratio = f'Reduction in number of stranded passengers after optim: {round(100 * ratio,2)}%'
        print(str_stranded_pax_ratio)
        file.write(str_stranded_pax_ratio + '\n')
        str_average_landing = f"average deviation for landing flights during disruption: {round(float(np.sum(list(delay_landing.values()))/nb_arr_flights_disruption), 1)}min"
        str_total_landing =f"total deviation for landing flights: {int(total_landing_delay)}min"
        str_nb_delayed_landing = f"nb of arrival flights that have been delayed: {np.sum(np.array(list(delay_landing.values())) !=0)}"
        print(str_average_landing)
        print(str_total_landing)
        print(str_nb_delayed_landing)
        file.write(str_average_landing + "\n")
        file.write(str_total_landing + "\n")
        file.write(str_nb_delayed_landing + "\n")
        str_average_takeoff =  f"average deviation for  departure flights: {round(float(np.sum(list(delay_takeoff.values()))/nb_dep_flights_disruption), 1)}min"
        str_total_takeoff =f"total deviation for departure flights: {int(total_takeoff_delay)}min"
        print(str_average_takeoff)
        print(str_total_takeoff)
        file.write(str_average_takeoff+'\n')
        file.write(str_total_takeoff+'\n')
        str_nb_delayed_takeoff = f"nb of departure flights that have been delayed: {np.sum(np.array(list(delay_takeoff.values())) !=0)}"
        print(str_nb_delayed_takeoff)
        file.write(str_nb_delayed_takeoff + "\n")
        str_slots = f"nb of departure flights that didn't respect their slot: {nb_slots_violated}"
        print(str_slots)
        file.write(str_slots + '\n')
        str_induced_delay = f'propagate delay after end of the disruption: {induced_delay}min'
        print(str_induced_delay)
        file.write(str_induced_delay+'\n')
    if test_constraint:
        is_runway_throughput_satisfied(airport, runway_ao, is_step_constraint=runway_step_constraint)
        is_taxi_capacity_satisfied(airport, final_taxi_congestion, is_step_constraint=taxi_step_constraint)
        is_terminal_capacity_satisfied(airport, final_terminals_congestion,
                                       is_step_constraint=terminal_step_constraint)
        is_air_connecting_pax_satisfied(flightset, passengers)
        is_turn_around_constraint_satisfied(flightset, airport)
        is_slot_constraint_satisfied(flightset)
    if display_constraints:
        display_throughput_constraints(airport, runway_bo,runway_ao,
                                       is_step_constraint=runway_step_constraint,
                                       savefig=savefig)
        display_capacity_constraint(airport, init_terminals_congestion, final_terminals_congestion,
                                    init_taxi_congestion, final_taxi_congestion,
                                    is_taxi_step_constraint=taxi_step_constraint,
                                    is_terminal_step_constraint=terminal_step_constraint,
                                    savefig=savefig)

    display_slot_violation(flightset,savefig=savefig)
    display_buffer_transfer_air_pax(flightset,passengers,savefig=savefig)
    display_barplot_delay(flightset,savefig=savefig)
    display_hourly_stranded_flight_delay(init_missed_pax,final_missed_pax,flightset,delay_landing,delay_takeoff, savefig)
    return nb_stranded_pax_bo, nb_stranded_pax_ao, total_landing_delay, total_takeoff_delay, nb_slots_violated, \









def plot_delay_model(t_min,t_max,t_start,t_end,start_delay):
    x =np.array([t_min,t_start+2,t_start+2,t_end+2,t_end+2,t_max])
    y=np.array([0,0,start_delay,0,0,0])
    y_train=[0,0,start_delay,start_delay,0,0]
    fig,ax=plt.subplots(figsize=(12,6))
    ax.plot(x,y,label='Road/Subway')
    ax.plot(x,y_train, label='Train',alpha=0.7)
    xticks = np.arange(4,26,2)
    ax.set_xticks(xticks)
    start_point = (t_start, 20)
    end_point = (t_start+2, 20)
    # Draw the arrow using annotate
    ax.annotate('', xy=end_point, xytext=start_point,
                arrowprops=dict(arrowstyle='<->', linewidth=2, color='red'))
    ax.text((start_point[0]+end_point[0])/2, end_point[1]+2, 'Access time', ha='center', va='bottom')
    ax.axvline(t_start,linestyle='--',color='red',label='Start/End disruption time')
    ax.axvline(t_end,linestyle='--',color='red')
    ax.set_xticklabels([str(e)+'h' for e in xticks])
    ax.set_ylabel('Passenger delay (minutes)',fontsize=12)
    ax.legend(fontsize=12,loc=1)
    plt.savefig("Results/delayModelling.png",bbox_inches='tight')

